"""Provide a process pool with enhanced performance on Windows."""


import multiprocessing
import os
import pickle
from pathlib import Path
from platform import system
from tempfile import mkstemp
from types import TracebackType
from typing import Any, Callable, Optional, Tuple, Type
import logging
import time

logger = logging.getLogger("pycomo")
logger.info('Process Pool Logger initialized.')


__all__ = ("ProcessPool",)


def _init_win_worker(filename: str) -> None:
    """Retrieve worker initialization code from a pickle file and call it."""
    logger.debug(f"Running _init_win_worker")
    with open(filename, mode="rb") as handle:
        func, *args = pickle.load(handle)
    func(*args)
    logger.debug("Finished _init_win_worker")


class ProcessPool:
    """Define a process pool that handles the Windows platform specially."""

    def __init__(
        self,
        processes: Optional[int] = None,
        initializer: Optional[Callable] = None,
        initargs: Tuple = (),
        maxtasksperchild: Optional[int] = None,
        **kwargs
    ) -> None:
        """
        Initialize a process pool.

        Add a thin layer on top of the `multiprocessing.Pool` that, on Windows, passes
        initialization code to workers via a pickle file rather than directly. This is
        done to avoid a performance issue that exists on Windows. Please, also see the
        discussion [1_].

        References
        ----------
        .. [1] https://github.com/opencobra/cobrapy/issues/997

        """
        logger.debug("Init ProcessPool")
        super().__init__(**kwargs)
        self._filename = None
        if initializer is not None and system() == "Windows":
            s_time = time.time()
            logger.debug(f"Process pool initialize")
            descriptor, self._filename = mkstemp(suffix=".pkl")
            # We use the file descriptor to the open file returned by `mkstemp` to
            # ensure that the resource is closed and can later be removed. Otherwise
            # Windows will cause a `PermissionError`.
            with os.fdopen(descriptor, mode="wb") as handle:
                pickle.dump((initializer,) + initargs, handle)
            initializer = _init_win_worker
            initargs = (self._filename,)
            logger.debug(f"Process pool init finished in {time.time() - s_time}")
        self._pool = multiprocessing.Pool(
            processes=processes,
            initializer=initializer,
            initargs=initargs,
            maxtasksperchild=maxtasksperchild,
        )
        logger.debug("Finished init ProcessPool")

    def __getattr__(self, name: str, **kwargs) -> Any:
        """Defer attribute access to the pool instance."""
        logger.debug(f"ProcessPool get {name}")
        return getattr(self._pool, name, **kwargs)


    def __enter__(self) -> "ProcessPool":
        """Enable context management."""
        logger.debug("ProcessPool enter context")
        self._pool.__enter__()
        return self


    def __exit__(
        self,
        exc_type: Optional[Type[BaseException]],
        exc_val: Optional[BaseException],
        exc_tb: Optional[TracebackType],
    ) -> Optional[bool]:
        """Clean up resources when leaving a context."""
        # The `multiprocessing.Pool.__exit__` only terminates pool processes. For a
        # clean exit, we close the pool and join the pool processes first.
        logger.debug("ProcessPool starting exit context")
        try:
            logger.debug("ProcessPool try to close")
            self._pool.terminate()
            logger.debug("ProcessPool try to join")
            self._pool.join()
            logger.debug("ProcessPool joined")
        finally:
            logger.debug("ProcessPool try cleanup")
            self._clean_up()
        logger.debug("ProcessPool try super exit")
        result = self._pool.__exit__(exc_type, exc_val, exc_tb)
        logger.debug("ProcessPool super exit done")
        return result


    def close(self) -> None:
        """
        Close the process pool.

        Prevent any more tasks from being submitted to the pool. Once all the tasks have
        been completed, the worker processes will exit.

        """
        logger.debug("ProcessPool starting to close")
        try:
            self._pool.close()
        finally:
            self._clean_up()
        logger.debug("ProcessPool finished closing")

    def _clean_up(self) -> None:
        """Remove the dump file if it exists."""
        logger.debug("ProcessPool starting to cleanup")
        if self._filename is not None and Path(self._filename).exists():
            Path(self._filename).unlink()
        logger.debug("ProcessPool cleanup finished")
